#!/bin/bash
set -euo pipefail

# error codes return
EXCODE_OK=0
EXCODE_PREREQS=$(( 1 << 0 ))
EXCODE_INPUT=$((1 << 1))
EXCODE_SPLIT=$(( 1 << 2 ))
EXCODE_BLAST=$(( 1 << 3 ))
EXCODE_SPARCLBL=$(( 1 << 4 ))


# BLAST batch files split from input files
BLAST_BATCH=blbatch

# define commandline switches
OPT_SGLFILE=-s
OPT_SGLFILE_L=--single_qfile

OPT_QUERIES=-q
OPT_QUERIES_L=--queries

OPT_PRECEDENCE=-p
OPT_PRECEDENCE_L=--precedence

OPT_DB=-b
OPT_DB_L=--db

OPT_DATA=-d
OPT_DATA_L=--datadir

OPT_OUT=-o
OPT_OUT_L=--outfile

OPT_BL_PFX=-bl_

# max number of BLAST processes. Default is 4
OPT_MAX_PROCS=-t
OPT_MAX_PROCS_L=--max-BLAST-procs

OPT_BATCH_SIZE=-z
OPT_BATCH_SIZE_L=--blast_batch_size


OPT_HELP=-h
OPT_HELP_L=--help



# mandatory BLAST parameters. 
BLASTCMD=('-parse_deflines' '-outfmt' '11' '-num_threads' '0')

# BLAST options not allowed to specify
X_BLASTCMD=('-parse_deflines' '-outfmt' '-num_threads')

##check against X_BLASTCMD
chk_x_blastcmd()
{
	local key=${1}
    local xk
	for xk in "${X_BLASTCMD[@]}"
	do
		if [[ ${key} == ${xk} ]]
		then
			return 0
		fi
	done
	return 1
}


BLASTDB_DEF="/netmnt/vast01/dd/cdd_ihp/SparcleLabels/cddsplit/current/cdd"
DEF_MAX_BL_PROCS=4
DEF_BL_BATCH_SZ=2048

DONE_FLG="rpsblast.done"
ABORT_FLG="rpsblast.abort"
###################################################################
#	Looking for binary in the same directory
###################################################################
script_name=$(basename "$0")
bindir=$(dirname "$0")
bindir=$(cd ${bindir}; pwd)
sparclbl_bin=${bindir}/sparclbl
sparclbl_ini=${bindir}/sparclbl.ini
rpsblast_bin=${bindir}/rpsblast
run_rpsb=${bindir}/runrpsb.sh
curr_dir=$(pwd)

#####################
##	parameters
#####################
##Single file process. ignore qf_mask
single_seqfile=

##Search using this mask
qf_mask=

##positional fasta files
qfiles=()

##precedence definition file
prec_def=

##sparcle data file location
datadir=

##how many threads. default is 0 -- automatically set according to number of files

##other arguments pass to BLAST
declare -A blargs
##customizable defaults
blargs["-evalue"]="0.01"


##output file
outfile=

num_procs=${DEF_MAX_BL_PROCS}
batch_size=${DEF_BL_BATCH_SZ}

usage()
{
	local xcode=0 
	
	if [[ $# -gt 0 ]]	##has error
	then
		local ec=1
		xcode=${EXCODE_PREREQS}
		echo "ERROR ########################################" 1>&2
		while (( $# ))
		do
			echo "# ${ec}:	$1" 1>&2
			shift
			(( ++ec )) || :
		done
		echo "##############################################" 1>&2
	fi

	cat <<EOF 1>&2
Performs RPSBlast and assigns a protein name to each sequence according to CDD functional annotations.

Usage:
    
    ${script_name} [options] <fasta file>[ fasta file[ ...]]
        -- Input FASTA files as positional arguments
    
    
    ${script_name} [options] ${OPT_QUERIES}|${OPT_QUERIES_L} "/path/to/*.fasta"
        -- Input multiple FASTA files as a glob. Note: Wildcards can only appear in basename.

    ${script_name} [options] ${OPT_SGLFILE}|${OPT_SGLFILE_L} "/path/to/specific_sequences.fasta"
        -- Specify a single input file. It is equivalent to specify with only one positional argument, but will ignore any other input files.
    
    ${script_name} [options]
        -- If no input specified, the script will read FASTA sequences from standard input.
	

Parameter description:
    ${OPT_PRECEDENCE}, ${OPT_PRECEDENCE_L} <precedence_def>
        Mandatory. Provide the file that defines the confidence level of each SPARCLE related naming source (CDD, CDDdom, CDDauto etc)
        Each line in the file should contain the name of name source, a delimiter (the tab character) and the associated confidence level:
        
            CDD	60
            CDDdom	5
            CDDauto	20
            .......
        
    ${OPT_OUT}, ${OPT_OUT_L} </path/to/output_file>
        Specify the output xml file. If omitted, output to stdout.
        
    ${OPT_DATA}, ${OPT_DATA_L} <data_dir>
        Specify the path to the directory that contains CDD annotation data files. If omitted, using the path defined in sparclbl.ini.
    
    ${OPT_DB}, ${OPT_DB_L} <path/to/db>
        Mandatory. Specify database for rpsblast.
    
    ${OPT_BL_PFX}xxx yyy
        argument passed to rpsblast as -xxx yyy. Not all options are passed though, some may need to have specific values.
        
    ${OPT_MAX_PROCS}, ${OPT_MAX_PROCS_L} n
        Maximal number of simultaneous BLAST processes. Default is ${DEF_MAX_BL_PROCS}
    
    ${OPT_BATCH_SIZE}, ${OPT_BATCH_SIZE_L} n
        Number of sequence for each rpsblast process. Default is ${DEF_BL_BATCH_SZ}
        
    ${OPT_HELP}, ${OPT_HELP_L}
        Display this help screen and ignore all other arguments.

EOF

exit $xcode
}

has_file_input=

while (( $# ))
do
	cmd="$1"
	shift
	case $cmd in
	${OPT_SGLFILE}=*|${OPT_SGLFILE_L}=*)	#single multientry fasta file as input. ignore others
		single_seqfile="${cmd#*=}"
        has_file_input=YES
		;;
	${OPT_SGLFILE}|${OPT_SGLFILE_L})	#single multientry fasta file as input. ignore others
		single_seqfile="${1}"
        has_file_input=YES
		shift
		;;
	${OPT_QUERIES}=*|${OPT_QUERIES_L}=*)
		qf_mask="${cmd#*=}"
        has_file_input=YES
		;;
	${OPT_QUERIES}|${OPT_QUERIES_L})
		qf_mask="${1}"
        has_file_input=YES
		shift
		;;
	${OPT_PRECEDENCE}=*|${OPT_PRECEDENCE_L}=*)
		prec_def="${cmd#*=}"
		;;
	${OPT_PRECEDENCE}|${OPT_PRECEDENCE_L})
		prec_def="${1}"
		shift
		;;
	${OPT_DB}=*|${OPT_DB_L}=*)
		blargs[-db]="${cmd#*=}"
		;;
	${OPT_DB}|${OPT_DB_L})
		blargs[-db]="${1}"
		shift
		;;
	${OPT_DATA}=*|${OPT_DATA_L}=*)
		datadir="${cmd#*=}"
		;;
	${OPT_DATA}|${OPT_DATA_L})
		datadir="${1}"
        shift
		;;
	${OPT_BL_PFX}*)
		k=${cmd%%=*}
		bk="-${k#*_}"
        v=
        
        if [[ ${k} == ${cmd} ]]
        then
            v="${1}"
            shift
        else
            v="${cmd#*=}"
        fi
        
		if ! chk_x_blastcmd ${bk} :
		then
            blargs[${bk}]="$v"
		fi
		;;
	${OPT_OUT}=*|${OPT_OUT_L}=*)
		outfile="${cmd#*=}"
		;;
	${OPT_OUT}|${OPT_OUT_L})
		outfile="${1}"
		shift
		;;
    ${OPT_MAX_PROCS}=*|${OPT_MAX_PROCS_L}=*)
        num_procs="${cmd#*=}"
        ;;
    ${OPT_MAX_PROCS}|${OPT_MAX_PROCS_L}) 
        num_procs="${1}"
        shift
        ;;
        
    ${OPT_BATCH_SIZE}=*|${OPT_BATCH_SIZE_L}=*)
        batch_size="${cmd#*=}"
        ;;
    ${OPT_BATCH_SIZE}|${OPT_BATCH_SIZE_L}) 
        batch_size="${1}"
        shift
        ;;    
    
	-h|--help)
		usage
		;;
    ##ignore list: 
	-m=*|-n=*)
	;;
    -m|-n)
    shift
    ;;
	##any unrecognized positional arguments are considered query files
	*)
        if [[ '-' != ${cmd:0:1} ]]
		then
			qfiles+=("${cmd}")
            has_file_input=YES
        else
            echo "Unknown command switch -> \"${cmd}\"" 1>&2
		fi
		;;
	esac
done

##verify input parameters
fatalerrs=0

if [[ ! -x ${sparclbl_bin} ]]
then
	echo "Binary executable ${sparclbl_bin} does not exist" 1>&2
	((++fatalerrs)) || :
fi

if [[ ! -x ${rpsblast_bin} ]]
then
	echo "Binary executable ${rpsblast_bin} does not exist" 1>&2
	((++fatalerrs)) || :
fi

if [[ ! -n ${prec_def} ]]
then
	echo "Mandatory parameter not specified: \"${OPT_PRECEDENCE} <precedence_def_file>\"" 1>&2
	((++fatalerrs)) || :
fi


if [[ fatalerrs -gt 0 ]]
then
	usage
	exit ${EXCODE_PREREQS}
fi

echo "$(date): SPARCLE-labeling process starting ..." 1>&2

if [[ -z "${datadir}" ]]
then
	datadir=()
else
	datadir=(-d "${datadir}")
fi

if [[ ! -n "${blargs['-db']-}" ]]
then
	blargs['-db']="${BLASTDB_DEF}"
fi

if [[ -z "${outfile}" ]]
then
	outfile=()
else
	outfile=(-o "${outfile}")
fi

if [[ num_procs -le 0 ]]
then
    num_procs=${DEF_MAX_BL_PROCS}
fi

if [[ batch_size -le 0 ]]
then
    batch_size=${DEF_BL_BATCH_SZ}
fi

blastargs=()
for blkey in ${!blargs[@]}
do
	echo "blastargs+=(${blkey} ${blargs[${blkey}]})"
    if [[ ! -n ${blargs[${blkey}]} ]]
    then
        blastargs+=("${blkey}") # flag
    else
        blastargs+=("${blkey}" "${blargs[${blkey}]}")
    fi
done

echo "\${blastargs[@]-}=${blastargs[@]-}" 1>&2


##########################################################
# avoid orphan processes, kill child processes upon exit
##########################################################
pid_sparclbl=
pids_rpsb=()


__onexit__()
{
    
    if [[ -n $pid_sparclbl && pid_sparclbl -gt 0 ]]
    then
        echo "Killing sparclbl process ${pid_sparclbl} ..." 1>&2
        kill -9 ${pid_sparclbl} || :
    fi
    
    local proc pid
    for proc in "${pids_rpsb[@]-}"
    do
        if [[ -n ${proc} ]]
        then
            pid=${proc%%\:*}
            if [[ -n "${pid}" && pid -ne 0 ]]
            then
                echo "Killing rpsblast process ${pid} ..." 1>&2
            kill -9 ${pid} || :
            fi
        fi
    done
    
}

trap __onexit__ EXIT


runtstamp=$(date +%Y%m%d_%H%M%S)
work_dir=${curr_dir}/${script_name}_${runtstamp}
qlist=${work_dir}/qlist
split=${work_dir}/split

if [[ ! -d ${work_dir} ]]
then
	mkdir -- ${work_dir}
else
	rm -rf -- ${work_dir}/*
fi

split_src=${work_dir}/split_src
blast_tmp=${work_dir}/blast_tmp
blast_done=${work_dir}/blast_done
sparc_work=${work_dir}/sparclbl_work
sparc_out=${work_dir}/sparclbl_out
log_dir=${work_dir}/logs

mkdir -- ${blast_tmp} ${blast_done} ${sparc_work} ${sparc_out} ${log_dir}

blast_fails=0

# - return 0 if submit success. return 1 if not (no available space)
do_blast()
{
    set +x
    local infile=
    
    if [[ $# -gt 0 ]]
    then
        infile="$1"
    fi
    
    local total_procs=${#pids_rpsb[@]}
    local pidx rc pid asnfile
    
    local first_avail=
    local total_running=0
    
    for ((pidx=0; pidx < total_procs; ++pidx))
    do
        if [[ -n ${pids_rpsb[pidx]} ]]
        then
            pid=${pids_rpsb[pidx]%%:*}
            asnfile=${pids_rpsb[pidx]#*:}
            orifile=${asnfile#*:}
            asnfile=${asnfile%:*}
            
            if ! kill -0 ${pid} 2>/dev/null :
            then
                wait ${pid} && rc=$? || rc=$?
                if [[ -f "${asnfile}" ]]
                then
                    mv "${asnfile}" "${sparc_work}/"
                fi
                if [[ $(basename ${orifile}) == ${BLAST_BATCH}* ]]
                then
                    rm -f "${orifile}" || :
                fi
                
                if [[ rc -eq 0 ]]
                then
                    echo "Blast process ${pid} completed successfully" 1>&2
                else
                    
                    echo "Blast process ${pid} failed with code ${rc}" 1>&2
                    ((++blast_fails)) || :
                fi
                    
                pids_rpsb[pidx]=
                if [[ ! -n ${first_avail} ]]
                then
                    first_avail=${pidx}
                fi
            else
                (( ++total_running )) || :
            fi
        elif [[ ! -n ${first_avail} ]]
        then
            first_avail=${pidx}
        fi
    done
    
    
    #set -x
    
    rc=0
    if [[ -n "$infile" ]]   # we need to launch it and return
    then
        if [[ -n ${first_avail} ]]  # has slot
        then
            asnfile=$(basename ${infile})
            asnfile="${blast_tmp}/${asnfile%.*}.asn"
            "${rpsblast_bin}" -query "${infile}" "${BLASTCMD[@]}" "${blastargs[@]}" -out "${asnfile}" &
            pid=$!
            
            pids_rpsb[${first_avail}]="${pid}:${asnfile}:${infile}"
            echo "Launched pids_rpsb[$first_avail]=${pids_rpsb[first_avail]}" 1>&2
            (( ++total_running )) || :
        else
            rc=1
        fi
        return ${rc}
    fi
        
    # if no submit file, consider done. wait for all BLAST processes to exit
    for ((pidx=0; pidx < total_procs; ++pidx))
    do
        if [[ -n ${pids_rpsb[pidx]} ]]
        then
            pid=${pids_rpsb[pidx]%%:*}
            asnfile=${pids_rpsb[pidx]#*:}
            orifile=${asnfile#*:}
            asnfile=${asnfile%:*}
            
            wait ${pid} && rc=$? || rc=$?
            if [[ -f "${asnfile}" ]]
            then
                mv "${asnfile}" "${sparc_work}/"
            fi
            if [[ $(basename ${orifile}) == ${BLAST_BATCH}* ]]
            then
                rm -f "${orifile}" || :
            fi
            
            if [[ rc -eq 0 ]]
            then
                echo "Blast process ${pid} completed successfully" 1>&2
            else
                
                echo "Blast process ${pid} failed with code ${rc}" 1>&2
                ((++blast_fails)) || :
            fi
                
            pids_rpsb[pidx]=
        fi
    done
    return 0
}

input_files=()
if [[ ! -n ${has_file_input} ]] # read from stdin
then
    echo "$(date +%Y%m%d_%H%M%S): Reading input from stdin ..." 1>&2
    input_files=("/proc/$$/fd/0")   # stdin
elif [[ -n ${single_seqfile} ]]
then
    if [[ ! -r ${single_seqfile} ]]
    then
        echo "Specified single input file does not exist or cannot read -> ${single_seqfile}" 1>&2
        exit $EXCODE_INPUT
    fi
    input_files=( ${single_seqfile} )
else
    if [[ -n "${qf_mask}" ]]
	then
		qf="${qf_mask##*/}"
		qfpath="${qf_mask%/*}"
        input_files=( $(find "${qfpath}" -maxdepth 1 -type f -name "${qf}") )
	fi
    
    if [[ ${#qfiles[@]} -gt 0 ]]
    then
        for afile in "${qfiles[@]}"
        do
            if [[ -f "${afile}" ]]
            then
                input_files+=( "${afile}" )
            else
                echo "File ${afile} does not exist" 1>&2
            fi
        done
    fi
    
    if [[ ${#input_files[@]} -eq 0 ]]
    then
        echo "None of the input files are readable, abort ..." 1>&2
        exit $EXCODE_INPUT
    fi
fi


excode=$EXCODE_OK
if [[ num_procs -eq 1 ]]    # single process, nothing fancy, just do files one by one
then
    rc=0
    if [[ ! -n ${has_file_input} ]] # read from stdin
    then
        asnfile="${sparc_work}/stdinput.asn"
        echo "\"${rpsblast_bin}\" -query - \"${BLASTCMD[@]-}\" \"${blastargs[@]-}\" -out \"${asnfile}\"" 1>&2
    
        "${rpsblast_bin}" -query - "${BLASTCMD[@]-}" "${blastargs[@]-}" -out "${asnfile}" && rc=$? || rc=$?
        
        if [[ rc -ne 0 ]]
        then
            echo "${rpsblast_bin} failed on stdin with code ${rc}" 1>&2
            exit $EXCODE_BLAST
        fi
    else
        for afile in "${input_files[@]}"
        do
            asnfile=$(basename ${afile})
            asnfile="${sparc_work}/${asnfile%\.*}.asn"
            "${rpsblast_bin}" -query "${afile}" "${BLASTCMD[@]-}" "${blastargs[@]-}" -out "${asnfile}" && rc=$? || rc=$?
            
            if [[ rc -ne 0 ]]
            then
                echo "${rpsblast_bin} failed on file ${afile} with code ${rc}" 1>&2
                excode=$EXCODE_BLAST
                break
            fi
        done
    fi
    
    if [[ EXCODE_OK -eq excode ]]
    then
        echo "$(date +%Y%m%d_%H%M%S): Done blasting, starting ${sparclbl_bin} processing ..." 1>&2
        
        touch "${work_dir}/${DONE_FLG}"
        
        echo "\"${sparclbl_bin}\" -p \"${prec_def}\" \"${datadir[@]-}\" -e ${blargs['-evalue']} -conffile \"$sparclbl_ini\" -x \"${work_dir}/${DONE_FLG}\" -b \"${work_dir}/${ABORT_FLG}\" -m \"${sparc_work}/*.asn\" -k \"${blast_done}\" \"${outfile[@]-}\"" 1>&2
        
        "${sparclbl_bin}" -p "${prec_def}" "${datadir[@]-}" -e ${blargs['-evalue']} -conffile "$sparclbl_ini" -x "${work_dir}/${DONE_FLG}" -b "${work_dir}/${ABORT_FLG}" -m "${sparc_work}/*.asn" -k "${blast_done}" "${outfile[@]-}" && rc=$? || rc=$?
        
        if [[ rc -ne 0 ]]
        then
            echo "${sparclbl_bin} failed with code $rc" 1>&2
            excode=$EXCODE_SPARCLBL
        fi
    fi
    
else
    
    # echo "Launching splitting process" 1>&2
    
    # split_fa_files "${split_src}" "${input_files[@]}" &
    # split_pid=$!
    
    echo "Launching Sparclbl in background mode ..." 1>&2
    
    echo "\"${sparclbl_bin}\" -p \"${prec_def}\" \"${datadir[@]-}\" -e ${blargs['-evalue']} -conffile \"$sparclbl_ini\" -x \"${work_dir}/${DONE_FLG}\" -b \"${work_dir}/${ABORT_FLG}\" -m \"${sparc_work}/*.asn\" -k \"${blast_done}\" \"${outfile[@]-}\"" 1>&2
    
    "${sparclbl_bin}" -p "${prec_def}" "${datadir[@]-}" -e ${blargs['-evalue']} -conffile "$sparclbl_ini" -x "${work_dir}/${DONE_FLG}" -b "${work_dir}/${ABORT_FLG}" -m "${sparc_work}/*.asn" -k "${blast_done}" "${outfile[@]-}" &
    pid_sparclbl=$!
    
    # create batch BLAST work directory
    blast_work=${split_src}/blast_work
    mkdir -p -- "${blast_work}"
    
    # -initialize BLAST pid array
    for ((i=0; i<num_procs; ++i))
    do
        pids_rpsb+=('')
    done
    
    # split in main shell -- somehow subshell runs extremely slow in occasional cases
    
    mkdir -p -- "${split_src}"
    seq_count=0
    fidx=1
    
    waiting_fa=()
    for afile in "${input_files[@]}"
    do
        if [[ ! -r "${afile}" ]]
        then
            echo "Input file does not exist -- ${afile}" 1>&2
            continue
        fi
        
        while read aline || [[ -n ${aline} ]]
        do
            if [[ '>' == ${aline:0:1} ]]
            then
                if [[ seq_count -gt batch_size ]]
                then
                    if [[ -f "${split_src}/${BLAST_BATCH}${fidx}.fa" ]]
                    then
                        mv "${split_src}/${BLAST_BATCH}${fidx}.fa" "${blast_work}/"
                        waiting_fa+=("${blast_work}/${BLAST_BATCH}${fidx}.fa")
                        
                        if do_blast "${waiting_fa[0]}" :
                        then
                            waiting_fa=("${waiting_fa[@]:1}")
                        fi
                    fi
                    (( ++ fidx )) || :
                    :> "${split_src}/${BLAST_BATCH}${fidx}.fa"
                    seq_count=0
                fi
                (( ++seq_count )) || :
            fi
            printf "%s\n" "${aline}" >> "${split_src}/${BLAST_BATCH}${fidx}.fa"
        
        done < "${afile}"
    done
    
    echo "Done splitting input files. Total batches: ${fidx}" 1>&2
    
    # last file
    if [[ -f "${split_src}/${BLAST_BATCH}${fidx}.fa" ]]
    then
        mv "${split_src}/${BLAST_BATCH}${fidx}.fa" "${blast_work}/"
        waiting_fa+=("${blast_work}/${BLAST_BATCH}${fidx}.fa")
    fi
    
    while [[ ${#waiting_fa[@]} -gt 0 ]]
    do
        if do_blast "${waiting_fa[0]}" :
        then
            waiting_fa=("${waiting_fa[@]:1}")
        else
            # waiting for next available BLAST slot
            sleep 10
        fi
    done
    
    # wait for all BLAST to be done
    echo "All files submitted, waiting for BLAST to finish" 1>&2
    do_blast || :
    
    echo "All BLAST processes have finished" 1>&2
    
    if [[ blast_fails -gt 0 ]]
    then
        echo "${blast_fails} BLAST process(es) have failed" 1>&2
    fi
    
    # set done flat and wait for sparclbl to finish
    touch "${work_dir}/${DONE_FLG}"
    echo "Waiting for sparclbl binary to finish" 1>&2
    wait ${pid_sparclbl} && rc=$? || rc=$?
    pid_sparclbl=
    
    if [[ rc -eq 0 ]]
    then
        echo "${sparclbl_bin} completed successfully" 1>&2
    else
        echo "${sparclbl_bin} failed with code ${rc}" 1>&2
        (( excode |= EXCODE_SPARCLBL )) || :
    fi
fi

if [[ excode -ne 0 ]]
then
    echo "Sparclbl pipeline failed with code ${excode}, partial results may still be available" 1>&2
else
    echo "$(date): All done." 1>&2
fi


exit ${excode}   

